'''
Created on Jun 14, 2016

@author: alberts
'''

import os
import nibabel as nib
import numpy as np

########################################
########## Tissue segmentation #########
########################################

this_dir = os.path.dirname(__file__)
TISSUES = ['wm','gm','csf']

class TissueSegmenter():
    '''This class allows to segment a brain image into wm, gm and csf,
    based on an atlas registered to the brain image'''
    
    def __init__(self, 
                 image_path, 
                 mask_path,
                 atlas_paths,
                 max_iteration):
        ''' Constructor
        parameters
        -----------
            - image_path: string
                    path to image where to segment 
                    the brain tissues from
            - atlas_path: dictionary of key to string
                    keys are in tissues and map to 
                    paths to the corresponding tissue atlas
        instance variables
        ------------------
            - self.dim: tuple
                    dimension of the 3d images
            - self.mask: numpy ndarray (shape = self.dim)_
                    binary image with ones in the brain region
            - self.data: numpy array (shape = (brain_voxels,))
                    flattened input image containing only
                    intensities of the brain voxels
            - self.atlas: dictionary of numpy arrays (shape = (brain_voxels,))
                    dictionary mapping TISSUES to the flattened atlas maps,
                    containing only tissue probabilities of brain voxels
            - self.tissue_maps: dictionary of numpy arrays (shape = (brain_voxels,))
                    dictionary mapping TISSUES to the flattened probabilistic
                    tissue prediction of the brain_voxels
            - self.means and self.stds: dictionaries of floats
                    dictionary mapping TISSUES to the mean intensity and std 
                    of the gaussian corresponding to that tissue
        '''
        
        # load image and mask
        image = nib.load(image_path).get_data()
        self.dim = image.shape
        self.mask = nib.load(mask_path).get_data()
        
        # select voxels within the mask
        # Check if data is bool
        self.mask=np.array(self.mask,dtype=bool)
        # check if shape of mask equals shape of image
        if self.dim != self.mask.shape:   
            err = 'Mask should be of the same size'
            raise RuntimeError(err)
            
        self.data = image[self.mask]
        #self.data=np.nan_to_num(self.data)
        
        # load atlas and initialse tissue maps with the atlas
        self.atlas = {}
        self.tissue_maps = {}
        self.means = {}
        self.stds = {}
        self.affines={}
        
        for tissue in TISSUES:
            atlas_tmp = \
                nib.load(atlas_paths[tissue])
            #Create a dictionary of affine matrix of each volume, this is necesary for saving later
            self.affines[tissue]=atlas_tmp.affine
            atlas_tmp=atlas_tmp.get_data()
            # check if shape of atlas is conform with image shape
            if atlas_tmp.shape != self.dim:
                err = 'All input images should be in the same space'
                raise RuntimeError(err)
                
            self.atlas[tissue] =(atlas_tmp[self.mask])
            self.tissue_maps[tissue] = np.copy(self.atlas[tissue])
            self.means[tissue] = None
            self.stds[tissue] = None
            
        # normalize to make sure initialisation is as expected
        self._normalize(self.atlas)
        self._normalize(self.tissue_maps)
        
    def _normalize(self, tissue_dict):
        
        sum_maps = 0
        for tissue in TISSUES:
            sum_maps += tissue_dict[tissue]
            '''if not np.any(sum_maps):
                sum_maps = tissue_dict[tissue]
            else:
                sum_maps += tissue_dict[tissue]'''
        sum_maps[sum_maps==0] = np.finfo(float).eps
                
        for tissue in TISSUES:
            tissue_dict[tissue] =(tissue_dict[tissue]/sum_maps)        
            
    def run(self):
        '''Run the iterations'''
        
        for i in range(50):#(len(self.max_iteration)):
        
            self._e_step()
            self._m_step()
            
        return self.tissue_maps
            
    def _e_step(self):
        '''Update model parameters, i.e. the distributions
        modelling the intensities of the different tissues'''
        
        # Retrieve wm, gm and csf tissue maps
        # calculate intensity distributions for each tissue
        for tissue in TISSUES:
            self.means[tissue] = get_mean(self.tissue_maps[tissue],
                                          self.data)
            self.stds[tissue] = get_std(self.tissue_maps[tissue],
                                        self.data)
        
    def _m_step(self):
        '''Update the current predications, i.e. the tissue 
        probabilities for each tissue in the brain, depending
        on the model parameters calculated in the previous step
        and using the atlas tissue maps as a prior'''
        
        # calculate likelihood for each tissue map
        #likelihood = np.zeros_like(self.data)
        for tissue in TISSUES:
            # Use the atlas and the new distribution parameters to 
            # calculate how likeli a voxel is to belong to 
            # the corresponding tissue
            self.tissue_maps[tissue] = self.atlas[tissue] *\
                            get_posterior(self.data,
                                          self.means[tissue],
                                          self.stds[tissue])
            #likelihood += self.tissue_maps[tissue]
            
        #likelihood_value=np.sum(np.log(likelihood))        
        
        # normalise new tissue maps
        self._normalize(self.tissue_maps)
        
    def write_tissue_maps(self, paths_dict = None):
        '''Write the predicted tissue maps to 3d-volumes which
        can be opened in a visualisation tool'''
        
        if paths_dict is None:
            paths_dict = {}
            paths_dict['wm'] = \
                extend_basename(this_dir+'/data/','wm_flair')
            paths_dict['gm'] = \
                extend_basename(this_dir+'/data/','gm_flair')
            paths_dict['csf'] = \
                extend_basename(this_dir+'/data/','csf_flair')
                
        for tissue in TISSUES:
            
            arr = np.zeros(shape = self.dim)
            arr[self.mask] = self.tissue_maps[tissue]
            img = nib.Nifti1Image(arr,self.affines[tissue])
            nib.save(img, paths_dict[tissue])
            
            print 'Image saved at '+paths_dict[tissue]
            
        return paths_dict
            
##########################
########## Tools #########
##########################

def get_mean(weights, data):
    '''Function returning the mean of data,
    taking weights into account'''
    
    weight_data=weights*data
    weight_mean=np.sum(weight_data)/np.sum(weights)
    
    return weight_mean
    
def get_std(weights, data):
    '''Function returning the std of data,
    taking weights into account'''
    
    weightedMean = get_mean(weights, data)
    demeanedData = np.square(data - weightedMean) # data after subtraction of the mean and squared
    stdTissue = np.sqrt( np.sum(weights*demeanedData)/(np.sum(weights)))
    return stdTissue

def get_posterior(data, mean, std):
    ''' Calculate for each entry in data how likeli it is to 
    belong to a gaussian with specified mean and std'''

    mahalanobis_distance=(((data-mean)**2)/(std)**2)
    prob=np.exp(-mahalanobis_distance/2)/(((2*np.pi)**(1/2))*std)
    return prob
            
########################################
############# Set datapaths ############
########################################
        
def load_datapaths():
    '''Get image_paths and atlas_paths which can be 
    supplied to the TissueSegmenter '''

    this_dir = os.path.dirname(__file__)
   
    
    image_paths = {}
    image_paths['t1'] = os.path.join(this_dir, 
                                     'data', 
                                     't1_brain_extraction.nii.gz')
    image_paths['flair'] = os.path.join(this_dir, 
                                        'data', 
                                        'affine_registration_flair.nii')

    mask_path=os.path.join(this_dir,'data',
                                    'mask_trans',
                                    'flair_brain_extraction_mask.nii.gz')
    
    atlas_paths = {}
    atlas_paths['wm'] = os.path.join(this_dir, 
                                     'data', 
                                     't1_reg_white.nii')
    atlas_paths['gm'] = os.path.join(this_dir, 
                                     'data', 
                                     't1_reg_grey.nii')
    atlas_paths['csf'] = os.path.join(this_dir, 
                                      'data', 
                                      't1_reg_csf.nii')
    
    return image_paths,mask_path, atlas_paths

def extend_basename(path,extension):
    '''Extend a path with a string, e.g. to write 
    results in recongizable paths'''

    noExt_ext = os.path.splitext(path)
    new_filename = noExt_ext[0]+'_'+extension+noExt_ext[1]
    
    return new_filename

########################################
##### When this module is executed #####
########################################

if __name__ == "__main__":
    
    image_paths, mask_path,atlas_paths = load_datapaths()
    
    # Initialise segmenter
    segmenter = TissueSegmenter(image_paths['flair'],mask_path,
                                atlas_paths,40)
    
    # Run the algorithm
    segmenter.run()
    
    # Write results
    segmenter.write_tissue_maps()
        
